/*    Filename: async_handler.c
 *      Author: Darren Hart <dvhltc@us.ibm.com>
 * Description: Measure the latency involved with asynchronous event handlers.
 *              Specifically it measures the latency of the pthread_cond_signal
 *              call until the signalled thread is scheduled.
 * Compilation: gcc -O2 -g -D_GNU_SOURCE -I/usr/include/nptl -I ../../include  -L/usr/lib/nptl -lpthread -lrt -lm async_handler.c -o async_handler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (C) IBM Corporation, 2006
 *
 * 2006-Oct-20:	Initial version by Darren Hart <dvhltc@us.ibm.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "librt.h"
#include "libjvmsim.h"
#include "libstats.h"

#define SIGNAL_PRIO 89
#define HANDLER_PRIO 89
#define ITERATIONS 1000000
#define HIST_BUCKETS 100
#define PASS_US 100

nsec_t start;
nsec_t end;

#define CHILD_START   0
#define CHILD_WAIT    1
#define CHILD_HANDLED 2
#define CHILD_QUIT    3
atomic_t step;

pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t mutex;

void *signal_thread(void *arg)
{
	struct thread *t = (struct thread *)arg;

	while (atomic_get(&step) != CHILD_QUIT) {
		pthread_mutex_lock(&mutex);
		atomic_set(CHILD_WAIT, &step);
		if (pthread_cond_wait(&cond, &mutex) != 0) {
			perror("pthead_cond_wait");
			break;
		}
		end = rt_gettime();
		atomic_set(CHILD_HANDLED, &step);
		pthread_mutex_unlock(&mutex);
		while (atomic_get(&step) == CHILD_HANDLED)
			usleep(10);
	}
	printf("handler thread exiting\n");
}

void *handler_thread(void *arg)
{
	struct thread *t = (struct thread *)arg;

	int i;
	long delta, max, min;
	stats_container_t dat;
	stats_container_t hist;

	stats_container_init(&dat, ITERATIONS);
	stats_container_init(&hist, HIST_BUCKETS);

	min = max = 0;
	for (i = 0; i < ITERATIONS; i++) {
		/* wait for child to wait on cond, then signal the event */
		while (atomic_get(&step) != CHILD_WAIT)
			usleep(10);
		pthread_mutex_lock(&mutex);
		start = rt_gettime();
		pthread_mutex_unlock(&mutex);
		if (pthread_cond_signal(&cond) != 0) {
			perror("pthread_cond_signal");
			atomic_set(CHILD_QUIT, &step);
			break;
		}

		/* wait for the event handler to schedule */
		while (atomic_get(&step) != CHILD_HANDLED)
			usleep(10);
		pthread_mutex_lock(&mutex);
		delta = (long)((end - start)/NS_PER_US);
		dat.records[i].x = i;
		dat.records[i].y = delta;
		if (i == 0) 
			min = max = delta;
		else {
			min = MIN(min, delta);
			max = MAX(max, delta);
		}
		atomic_set((i == ITERATIONS-1) ? CHILD_QUIT : CHILD_START, &step);
		pthread_mutex_unlock(&mutex);
	}
	printf("recording statistics...\n");
	printf("Minimum: %d\n", min);
	printf("Maximum: %d\n", max);
	printf("Average: %f\n", stats_avg(&dat));
	printf("Standard Deviation: %f\n", stats_stddev(&dat));
	stats_hist(&hist, &dat);
	stats_container_save("samples", "Asynchronous Event Handling Latency Scatter Plot",\
			     "Iteration", "Latency (us)", &dat, "points");
	stats_container_save("hist", "Asynchronous Event Handling Latency Histogram",\
			     "Iteration", "Latency (us)", &hist, "steps");
	printf("signal thread exiting\n");
}

int main(int argc, char *argv[])
{
	int signal_id, handler_id;

	printf("-------------------------------\n");
	printf("Asynchronous Event Handling Latency\n");
	printf("-------------------------------\n\n");
	printf("Running %d iterations\n", ITERATIONS);

	rt_init(NULL, NULL, argc, argv);

	init_pi_mutex(&mutex);

	atomic_set(CHILD_START, &step);
	jvmsim_init();
	handler_id = create_fifo_thread(handler_thread, (void*)0, HANDLER_PRIO);
	signal_id = create_fifo_thread(signal_thread, (void*)0, SIGNAL_PRIO);

	join_thread(signal_id);
	join_threads();

	return 0;
}
